-------------------------------------------------------------------------------
-- Copyright (c) 2015 Sierra Wireless. All rights reserved.
--
-- Contributors:
--     Randall Pfeifer - initial implementation
--     Tom Ho - extended to discover/report proprietary vehicle MSCIIDs (AMMER-184)
--
--	Main application module.
-------------------------------------------------------------------------------

-- ----------------------------------------------------------------------------
-- REQUIRES
-- ----------------------------------------------------------------------------

local sched         = require "sched"
local log           = require "log"
local devicetree    = require "devicetree"
local airvantage    = require "airvantage"
local persist       = require "persist"
local timer         = require "timer"
local serial        = require 'serial'
local systime       = require "sched.timer.core"
local os            = require "os"
local psignal       = require "posixsignal"
local string        = require "string"
local bit           = require "bit32"
local pkginfo       = require "pkginfo"
local lfs           = require 'lfs'
local json          = require 'json'

-- application specific code module constructors
local _bbem = require 'bbem'
local _BnB =  require 'BnB'
local _m3da = require 'm3da'
local _telemetry = require 'telemetry'

-- Primary code objects
local transport
local BnB
local telemetry
local wanstate = require 'wanstate'

-- ----------------------------------------------------------------------------
-- CONSTANTS
-- ----------------------------------------------------------------------------

-- Set to false to disable debugging interface --
local DEBUG = true

local APP_NAME = "AMMER"
local DEVICE_ID = ""
local CONFIG_ASSET_NAME = "ammer_config"

-- ----------------------------------------------------------------------------
-- VARIABLES
-- ----------------------------------------------------------------------------

-- support for discover/reporting of telemetry cfg file proprietary MSCIIDs
local TELEM_CFG_FILE_TITLE    = nil -- Store Active Telemetry Cfg File's TITLE
local TELEM_CFG_FILE_REVISION = nil -- Store Active Telemetry Cfg File's REVISION
local TELEM_CFG_FILE_SIGNAL_NAME_TBL = {} -- { aleos.<msciid>, <datum name>  } pair
local TELEM_CFG_FILE_SIGNAL_UNIT_TBL = {} -- { aleos.<msciid>, <datum units> } pair
local TELEM_CFG_FILE_SIGNAL_REGISTERED_USERID_TBL = {} -- { aleos.<msciid>, <userid returned by devicetree.register()> } pair
local TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE = {} -- { <datum name>, <datum vaue> } pair; keep cache until time to publish
local TELEM_CFG_FILE_SIGNAL_LAST_PUBLISH  =  0 -- time at which TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE was last published to ALMS
local TELEM_CFG_FILE_SIGNAL_REPORT_PERIOD = 60 -- Must not publish more often that this period (seconds)

-- Scan Tool status
local TELEM_SCAN_TOOL_STATUS = nil -- 0 = Not present, 1 = Scan tool present causing delays, 2 = Scan tool present preventing TX

-- current ALEOS host software version
local SW_VERSION

-- set to true if CANBUS is available and enabled, else false
local CANBUS_ENABLED

-- These are set to point to either ignition input or I/O input in device tree
local IGNITION_SENSE_LINE
local IGNITION_SENSE_TIME

-- Cellular DPI
local ALEOS_CELLLINK_DPI = 0x30200000 -- 807403520

-- validators for common parameters
local function validate_posnum(pname, pvalue)
	if tonumber(pvalue) and tonumber(pvalue) >= 0 then
		return true
	end
	return false
end

-- Table to describe system options.
-- Each entry names the system option (accessed via M3DA, same as telemetry options),
-- and is itelf a table that describes the option. It's fields are as follows:
--	value		- Default value for the parameter. This value is used to initialize an entry in telemetry.options
--				  (which is persisted) upon startup if it doesn't exist. Existence of these values is used to
--				  indicate to need to read when all values are read.
--				  A value of false indicates that this parameter is external to telemetry.options table,
--				  is accessed entirely via handler functions, and is not initialized.
--	validate	- Function to indicate that the supplied new value is acceptable (return true) or not (return false).
--				  External parameters (value = false) are not altered if validation fails; others are set to default on
--				  validation failure.
--	set_handler	- Function called (if present) with param name and new value when setting value.
--				  The new value is also saved to (persisted) telemetry.options table if value setting != false (see above),
--				  whether or not there is a set_handler.
--	get_handler	- Function called (if present) with param name to return the value of the parameter upon read.
--				  If not present, the read value is taken from telemetry.options table.
--
-- Extra entries can be added to each definition for use by functions. Create table initially as temp cso table as trick
-- to allow self-referential table entries.
local default_system_options = {}
local cso = {
	cfg_version = {
		value = false,
		validate = function (pname, pvalue) return true end,
		-- Always return the app version info from the package
		get_handler = 
			function(pname)
				return pkginfo.manifest.components[1].version
			end,
		-- Ignore new incoming value, but log it. Probably never happen.
		set_handler =
			function(pname, pvalue)
				log(APP_NAME, "WARNING", "Attempt to configure cfg_version: " .. pvalue)
			end
	},
	sys_log_level = {
		value = "INFO",
		validate =
			function(pname, pvalue)
				local valid = {NONE=true,ERROR=true,WARNING=true,INFO=true,DETAIL=true,DEBUG=true,ALL=true}

				return valid[pvalue]
			end,
		set_handler =
			function(pname, pvalue)
				log.setlevel(pvalue, "AMMER")
			end
	},
	sys_tx_persist_rate = {
		value = 30,
		validate = validate_posnum,
		set_handler =
			function(pname, pvalue)
				transport:configure {persistrate = pvalue}
			end
	},
	sys_tx_persist_threshold = {
		value = 5,
		validate = validate_posnum,
		set_handler =
			function(pname, pvalue)
				transport:configure {persistthresh = pvalue}
			end
	},
	sys_tx_latency = {
		value = 5,
		validate = validate_posnum,
		set_handler =
			function(pname, pvalue)
				transport:configure {latency = pvalue}
			end
	},
	sys_tx_maxsend = {
		value = 256,
		validate = validate_posnum,
		set_handler =
			function(pname, pvalue)
				transport:configure {maxsend = pvalue}
			end
	},
	sys_tx_maxqueue = {
		value = 10000,
		validate = validate_posnum,
		set_handler =
			function(pname, pvalue)
				transport:configure {maxqueue = pvalue}
			end
	},
	sys_tx_queuetrail = {
		value = 500,
		validate = validate_posnum,
		set_handler =
			function(pname, pvalue)
				transport:configure {queuetrail = pvalue}
			end
	},
	-- OpenVPN Tunnel controls read/write via MSCI, not config
	tun_enabled = {
		csmid = "aleos.10033",
		value = "1",
		validate = validate_posnum,
		set_handler = function(pname, pvalue) devicetree.set(default_system_options[pname].csmid, pvalue) end,
		get_handler = function(pname) return devicetree.get(default_system_options[pname].csmid) end,
	},
	rssi_filter_len = {
		value = 10,
		validate = validate_posnum,
	},
}
default_system_options = cso
cso = nil

-- Empty file created to indicate device boot has been processed
local bootflagfile = "/tmp/RA/ammerbootflag"

-- Table of used MSCI parameter definitions (from MSCIIDs.h)
local msciid = {
	STS_POWERIN								= "aleos.266",
	STS_BOARDTEMP							= "aleos.267",
	STS_CPU_LOAD							= "aleos.10253",
	STS_MODEM_SENT_PERSIST					= "aleos.10283",
	STS_MODEM_RECV_PERSIST					= "aleos.10284",
	CFG_LOW_PWR_IGNITION_TIMER				= "aleos.11131",
	CFG_LOW_PWR_IO_LINE_TIMER				= "aleos.11143",
	STS_TIME_VALID							= "aleos.55010",
	CFG_DRECKON_ENABLE						= "aleos.10993",
	MP70_VEH_BEHAVE_EN						= "aleos.10995",
	STS_DRECKON_SENSOR_CALIBRATED			= "aleos.10996",
	CFG_TELEM_CANBUS_ENABLED				= "aleos.59077",
	CFG_TELEM_IGN_SOURCE					= "aleos.59076",
	GROUP_CANBUS							= "aleos.10312",
	RAP_ID									= "aleos.10928",
	CFG_TELEM_CANBUS_ENABLED_IGN_SOURCE		= "aleos.59080",
	CFG_AAF_RESERVE_SERIAL_PORT				= "aleos.10120",
	IO_DIGITAL_IN1							= "aleos.851",
	IO_DIGITAL_IN2							= "aleos.852",
	IO_DIGITAL_IN3							= "aleos.853",
	IO_DIGITAL_IN4							= "aleos.854",
	IO_DIGITAL_IN5							= "aleos.867",
	IO_DIGITAL_IN6							= "aleos.868",
	IO_ANALOG_IN1							= "aleos.855",
	IO_ANALOG_IN2							= "aleos.856",
	IO_ANALOG_IN3							= "aleos.857",
	IO_ANALOG_IN4							= "aleos.858",
	IO_ANALOG_IN5							= "aleos.873",
	IO_ANALOG_IN6							= "aleos.874",
	IO_ANALOG_IN7							= "aleos.875",
	IO_ANALOG_IN8							= "aleos.876",
    STS_TELEM_CFG_FILE_TITLE                = "aleos.59390",
    STS_TELEM_CFG_FILE_REVISION             = "aleos.59391",
    STS_TELEM_SCAN_TOOL_STATUS              = "aleos.59389",
	VPN1_STATE								= "aleos.3176",
	VPN2_STATE								= "aleos.3205",
	VPN3_STATE								= "aleos.3231",
	VPN4_STATE								= "aleos.3257",
	VPN5_STATE								= "aleos.3283",
	STS_CELL_RAT							= "aleos.264",
	STS_CELL_BAND							= "aleos.671",
    STS_LTE_RSRQ							= "aleos.10209",
    STS_LTE_RSRP							= "aleos.10210",
    STS_LTE_SINR							= "aleos.10211",
    STS_LTE_RSSI							= "aleos.261",
}

-- DELS event map
local delseventmap  = {
	telemetry                  = {messageid = 0,   name = "Telemetry Data"},
	appinfo                    = {messageid = 1,   name = "App Info"},
	caninfo                    = {messageid = 1,   name = "CANbus availability"},
	rap_id                     = {messageid = 1,   name = "RAP ID"},
	input                      = {messageid = 1,   name = "General Purpose Digital Inputs"},
	input_an                   = {messageid = 1,   name = "General Purpose Analog Inputs"},
	custom_can_data            = {messageid = 2,   name = "Custom CANbus data"},
	startup                    = {messageid = 256, name = "Gateway Start Up"},
	heartbeat                  = {messageid = 270, name = "Gateway Heartbeat"},
	uncleanshutdown            = {messageid = 269, name = "Gateway Unclean Shutdown"},
	shutdown                   = {messageid = 257, name = "Gateway Shut Down"},
	tun_state                  = {messageid = 293, name = "Management Tunnel State"},
	tun_address                = {messageid = 294, name = "Management Tunnel Address"},
	wanlinktxbytes             = {messageid = 520, name = "WAN Link Tx Bytes"},
	wanlinkrxbytes             = {messageid = 521, name = "WAN Link Rx Bytes"},
	wan_sig_ext                = {messageid = 535, name = "WAN Extended Signal report"},
	cpu_load                   = {messageid = 273, name = "CPU Load Average"},
	watchdogreboot             = {messageid = 278, name = "Gateway Software Watchdog System Reboot Determination"},
	main_supply                = {messageid = 768, name = "Main Supply Voltage"},
	board_temperature          = {messageid = 770, name = "Mainboard Temperature"},
	gpslocation                = {messageid = 772, name = "Gateway Location Via Internal GPS"},
	gpsspeed                   = {messageid = 773, name = "Gateway Speed Via Internal GPS"},
	gpsfix                     = {messageid = 774, name = "Gateway Internal GPS Fix Status"},
	ignitionstate              = {messageid = 775, name = "Ignition State"},
	bootignitionstate          = {messageid = 775, name = "Boot Ignition State"},
	vpn_profile_id_map         = {messageid = 1280, name = "Gateway VPN Profile to ID Map"},
	vpn_tunnel_status          = {messageid = 1281, name = "Gateway Individual VPN Tunnel Status"},
	acceleration_incident_start = {messageid = 70144, name = "Vehicle Behaviour Acceleration Incident Start"},
	acceleration_incident_end  = {messageid = 70145, name = "Vehicle Behaviour Acceleration Incident End"},
	deceleration_incident_start = {messageid = 70146, name = "Vehicle Behaviour Deceleration Incident Start"},
	deceleration_incident_end  = {messageid = 70147, name = "Vehicle Behaviour Deceleration Incident End"},
	cornering_incident_start   = {messageid = 70148, name = "Vehicle Behaviour Cornering Incident Start"},
	cornering_incident_end     = {messageid = 70149, name = "Vehicle Behaviour Cornering Incident End"},
--	crash_incident_start       = {messageid = 70150, name = "Vehicle Behaviour Crash Incident Start"},
--	crash_incident_end         = {messageid = 70151, name = "Vehicle Behaviour Crash Incident End"},
	dreckon_sensor_calibration = {messageid = 70152, name = "Vehicle Behaviour Sensor calibration"},
	dreckon_enabled            = {messageid = 70153, name = "Dead Reckoning Enabled"},
}
-- Table to relate telemetry parameter by name to DELS ID and format
local DELSparams = {
	["vehicle_speed"]					= {id =  13, format = "I"},		-- MPH (int)
	["engine_speed"]					= {id =  12, format = "I"},		-- RPM (int)
	["throttle_position"]				= {id =  17, format = "I"},		-- % (int)
	["odometer"]						= {id =  97, format = "I"},		-- Miles (int)
	["fuel_level"]						= {id =  47, format = "I"},		-- % (int)
	["engine_coolant_temp"]				= {id =   5, format = "I"},		-- Deg F (int)
	["mil_status"]						= {id = 317, format = "I"},		-- 0 (off) | 1 (on)
	["battery_voltage"]					= {id =  83, format = "S"},		-- Volts (1 dec)
	["pto_status"]						= {id =  30, format = "I"},		-- 0 (off) | 1 (on)
	["seatbelt_fastened"]				= {id = 193, format = "I"},		-- 0 (fastened) | 1 (not fastened)
	["brake_switch_status"]				= {id = 341, format = "I"},		-- 0 (off) | 1 (on)
	["ambient_air_temperature"]			= {id =  70, format = "S"},		-- Deg C (1 dec)
	["trip_fuel_consumption"]			= {id = 201, format = "I"},		-- Gallons (int)
	["distance_since_dtc_cleared"]		= {id =  49, format = "I"},		-- Miles (int)
	["transmission_fluid_temperature"]	= {id = 145, format = "S"},		-- Deg F (1 dec)
	["engine_run_time"]					= {id =  31, format = "I"},		-- Seconds (int)
	["vin"]								= {id =   2, format = "S"},		-- string
	["dtcs"]							= {id =   1, format = "S"}		-- comma-separated strings
}

-- Table to associate VPN event with number
local VPNInfo = {
		[msciid.VPN1_STATE] = {num = 0, name = "VPN 1"},
		[msciid.VPN2_STATE] = {num = 1, name = "VPN 2"},
		[msciid.VPN3_STATE] = {num = 2, name = "VPN 3"},
		[msciid.VPN4_STATE] = {num = 3, name = "VPN 4"},
		[msciid.VPN5_STATE] = {num = 4, name = "VPN 5"},
}
-- Table to associate VPN state with AMM state
local VPNStates = {
	["Not Enabled"]					= 3,
	["Disabled"]					= 3,
	["Connected"]					= 2,
	["Not Connected"]				= 3,
	["Error Connecting"]			= 3,
	["Error in Gateway"]			= 3,
	["Error in IKE Proposal(s)"]	= 3,
	["Error in ESP Proposal(s)"]	= 3,

	-- The full set of states supported by AMM.
	-- ALEOS does not support most of these.
	["-standby-"]					= 0,
	["-initial-connecting-"]		= 1,
	["-connected-"]					= 2,
	["-disconnected-"]				= 3,
	["-connecting-"]				= 4,
	
	-- use this if don't reconnize any others
	["default"]						= 3
}

-- ----------------------------------------------------------------------------
-- HELPER FUNCTIONS
-- ----------------------------------------------------------------------------

-- encode GPS coord into 9.23 format
local function encodeGps(coord)
	-- This nasty little hack is required because bit32 doesn't handle 32 bits
	if coord < 0 then
		-- Remember negate is complement/increment (hence the "1" and MSB)
		return (bit.bxor(math.floor((-coord) * (2 ^ 23) + 0.5), 0x7fffffff) + 0x80000001)
	else
		return (math.floor(coord * (2 ^ 23) + 0.5))
	end
end

-- Debugging routine to display bytes in a string
function showdata(str)
	if type(str) ~= "string" then
		return ""
	end

	local s = ""
	local i = 1
	while(string.byte(str,i)) do
		s = s..(string.format("%02x ",string.byte(str,i)))
		i = i + 1
	end
	return s
end

------------------------------------------------------------------
-- recieve_config - AirVantage data handler callback to recieve data 
-- 
-- Inputs:
-- 	assetInstance	- Asset Instance
-- 	value			- Value of recieved data
-- 	path			- data model path
--
-- Outputs:
-- 	telemetry.options
--
-- Returns:
-- 	"ok" to indicate to AirVantage that data has been handled
--
-- Notes
--	Recieves configuration data from AirVantage system and saves to
--	configuration parameters in telemetry opbject. Updates persistent
--	object store with new data.
------------------------------------------------------------------
local function recieve_config(assetInstance, value, path)
    --print("*** Recieved: ", assetInstance, value, path)
    local err = nil
    for pname,pvalue in pairs(value) do 

    	if type(pvalue) ~= "string" then
    		-- Sometimes we get wrong data type (function) despite specifying string. Fix it.
    		pvalue = ""
    	end
       	log(APP_NAME, "INFO", 'Configure %s="%s"', pname, tostring(pvalue))

		-- process as a system option if it's in defaults table
		-- system option value are all strings
		if default_system_options[pname] then
			if not default_system_options[pname].validate(pname, pvalue) then
				-- empty or invalid value specified - use default
				log(APP_NAME, "WARNING", "Invalid value for system option %s: %s", pname, tostring(pvalue))
				if default_system_options[pname].value then
					-- only update value if there's a default value
					telemetry.options[pname] = default_system_options[pname].value
					if default_system_options[pname].set_handler then
						default_system_options[pname].set_handler(pname, pvalue)
					end
				end
			else
				if default_system_options[pname].value then
					-- update value in options if default value is true (not false)
					-- some parameters don't keep values there (eg read/write dev tree)
					-- these don't need default values
					telemetry.options[pname] = pvalue
				end
				if default_system_options[pname].set_handler then
					default_system_options[pname].set_handler(pname, pvalue)
				end
			end
		else
			err = BnB:set_option(telemetry.options, pname, pvalue)
			if err ~= "ok" then
				log(APP_NAME, "ERROR", "Configure error (%s) %s=%s", err, pname, tostring(pvalue))
				-- return err
			end
		end
    end

	persist.save(CONFIG_ASSET_NAME, telemetry.options)
    telemetry:reconfigure()
    return "ok"
end

-- Get configuration as a table of strings, suitable for
-- sending to AMM (table of value=setting).
local function get_config(assetInstance, value, path)
	local options = {}
	local pnames = {
		{ext="enable", int="en"},
		{ext="threshold", int="th"},
		{ext="min", int="mn"},
		{ext="max", int="mx"}
	}

	for k,v in pairs(telemetry.options) do
		if default_system_options[k] then
			-- if param is in system defaults then it's a (simple) system setting
			local tv = v
			if default_system_options[k].get_handler then
				-- Read via handler if defined, else just use setting value
				tv = default_system_options[k].get_handler(k)
			end
			if tv then
				-- Only send parameters that have values
				-- (probably only MSCI on unsupported platform don't)
				options[k] = tostring(tv)
			end
		else
			-- it's a compound telemetry setting (enable, threshold, min, max)
			local PIDInfo = telemetry.streamer.PIDInfoByName[k]
			if PIDInfo then	-- make sure it's a valid option
				local opts = ""
				for i = 1, #pnames do
					if not PIDInfo.mask or not PIDInfo.mask[pnames[i].int] then
						-- hide masked sub-params according to master table
						opt = pnames[i].ext .. '=' .. v[pnames[i].int]
						if opts == "" then
							opts = opt
						else
							opts = opts .. "," .. opt
						end
					end
				end
				options[k] = opts
			end
		end
	end
	return options
end

-- airvantage handler callback to send the entire configuration to AMM.
-- Ignores any incoming parameters.
local function request_config(assetInstance, value, path)
	log(APP_NAME, "INFO", "request_config: path=" .. tostring(path) .. " value=" .. tostring(value))
	local options = get_config(assetInstance, value, path)
	assert(assetInstance:pushData("uplink", options))
	return "ok"
end

-- ----------------------------------------------------------------------------
-- DELS Event Generation Functions (from eventreporter app)
-- ----------------------------------------------------------------------------
-- Table to associate MP70 accelerometer incidents with dev tree params
-- start events and end events share structure, so group them
local vbistartparms = {
	acceleration_incident_start = "system.vehicle.accel.event_start",
	deceleration_incident_start = "system.vehicle.brake.event_start",
	cornering_incident_start    = "system.vehicle.side.event_start"
--	crash_incident_start = "system.vehicle.crash.event_start"
}
local vbiendparms = {
	acceleration_incident_end = "system.vehicle.accel.event_end",
	deceleration_incident_end = "system.vehicle.brake.event_end",
	cornering_incident_end    = "system.vehicle.side.event_end"
--	crash_incident_end = "system.vehicle.crash.event_end"
}

--[[
-- Until params are available by name ...
-- (delete these when MP70 is updated)
local vbistartparms = {
	acceleration_incident_start = "aleos.10985",
	deceleration_incident_start = "aleos.10987",
	ing_incident_start = "aleos.10989",
	crash_incident_start = "aleos.10991"
}
local vbiendparms = {
	acceleration_incident_end = "aleos.10986",
	deceleration_incident_end = "aleos.10988",
	cornering_incident_end = "aleos.10990",
	crash_incident_end = "aleos.10992"
}
--]]

-- Split a (comma) separated string into an array of fields
local function split(inputstr, sep)
	if sep == nil then
		sep = ","
	end
	local t={} ; i=1
	for str in string.gmatch(inputstr, "([^"..sep.."]+)") do
		t[i] = str
		i = i + 1
	end
	return t
end

-- Collect data for event t and generate an event from it. Some events
-- also apply the telemetry config filter to suppress unwanted events.
-- Some events also use v parameter, which contains the devicetree data
-- that triggered the event. Not all events use this (though more probably
-- should). Note that calling devicetree.get() can cause AAF to yield,
-- possibly causing unexpected results.

local dBm		-- rolling rssi average for wan_sig_ext
local function getevent(t, v)
    local event = {}
	event.messageid = delseventmap[t].messageid
	if t == "startup" then
		event.timestamp = 0		-- flag to timestamp as initial boot time
		event.param0 = 1		-- boot cause (not available, so use ignition)
		event.param1 = (tonumber(devicetree.get(msciid.STS_POWERIN)) or 0) * 1000
		event.extra = "ESN="..devicetree.get("system.hw_info.global_id")..
						",platform="..devicetree.get("system.hw_info.product_str") ..
						",board="..devicetree.get("system.hw_info.hw_ver")

	elseif t == "appinfo" then
		event.extra = "AMMER Ver = " .. pkginfo.manifest.components[1].version ..
				      ", build = " .. pkginfo.build ..
				      ", commit = " .. pkginfo.commit
		log(APP_NAME, "INFO", event.extra)

	elseif t == "caninfo" then
		if devicetree.get(msciid.GROUP_CANBUS) == "1" then
			if CANBUS_ENABLED then
				event.extra = "<CANBus>=<Enabled>"
			else
				event.extra = "<CANBus>=<Disabled>"
			end
		else
			event.extra = "<CANBus>=<N/A>"
		end
		log(APP_NAME, "INFO", event.extra)

	elseif t == "rap_id" then
		local rap = tostring(devicetree.get(msciid.RAP_ID))
		event.extra = string.format("<RAP_ID>=<%s>", rap)
		local tmp = {[t]=rap}
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end

	elseif t == "wan_sig_ext" then
		if string.match(tostring(devicetree.get(msciid.STS_CELL_RAT)), "LTE") then
			-- event only valid for LTE
			local rssi = tonumber(devicetree.get(msciid.STS_LTE_RSSI))
			if dBm == nil then
				dBm = rssi
			else
				dBm = dBm - (dBm / telemetry.options.rssi_filter_len) + (rssi / telemetry.options.rssi_filter_len)
			end
			local tmp = {[t]=dBm}
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			else
				local rsrq = tonumber(devicetree.get(msciid.STS_LTE_RSRQ))
				local rsrp = tonumber(devicetree.get(msciid.STS_LTE_RSRP))
				local sinr = tonumber(devicetree.get(msciid.STS_LTE_SINR))
				local band = tonumber(string.match(devicetree.get(msciid.STS_CELL_BAND), "%d"))
				event.param0 = ALEOS_CELLLINK_DPI
				-- This format doesn't match DELS docs, but does match MG90
				event.extra = string.format("dBm=%d,BN=%d,BW=,RSSI=%d,RSRP=%d,RSRQ=%.1f,SINR=%.1f", dBm, band, rssi, rsrp, rsrq, sinr)
			end
		else
			event = nil
		end

	elseif t == "cpu_load" then
		local cpu = devicetree.get(msciid.STS_CPU_LOAD)
		if cpu then
			event.param0 = tonumber(cpu) * 100
			event.param1 = 900 -- observation period (sec) - STS_CPU_LOAD gets 15 min avg
			local tmp = {[t]=event.param0}
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		else
			event = nil
		end

	elseif t == "main_supply" then
		local supply = devicetree.get(msciid.STS_POWERIN)
		if supply then
			event.param0 = tonumber(supply) * 1000
			local tmp = {[t]=event.param0}
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		else
			event = nil
		end

	elseif t == "board_temperature" then
		local btemp = devicetree.get(msciid.STS_BOARDTEMP)
		if btemp then
			event.param0 = tonumber(btemp)
			local tmp = {[t]=event.param0}
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		else
			event = nil
		end

	elseif t == "ignitionstate" or t == "bootignitionstate" then
		if t == "bootignitionstate" then
			event.timestamp = 0		-- flag to timestamp as initial boot time
		end

		local ignstate

		if IGNITION_SENSE_LINE == "always_on" then
			ignstate = 1
		else
			ignstate = devicetree.get(IGNITION_SENSE_LINE)
		end
		event.param0 = ignstate
		event.param1 = tonumber(devicetree.get(IGNITION_SENSE_TIME))
		
		local tmp = {["ignition_state"] = ignstate}
		telemetry:filter(tmp)
		if tmp["ignition_state"] == nil then
			event = nil
		end

	elseif t == "heartbeat" then
		local tmp = { [t] = 0 } -- no threshold to filter on
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		else
			persist.save("lastheartbeat", transport:systime() * 1000)
			event.param0 = math.floor(systime.time())
		end

	elseif t == "uncleanshutdown" then
		local lasthb = persist.load("lastheartbeat")
		if type(lasthb) ~= "number" or lasthb < transport.reftime then
			event = nil
			log(APP_NAME, "INFO", "Unclean shutdown event suppressed - invalid timestamp")
		else
			-- work-around for brain-dead bit32 functions
			event.param0 = math.floor(lasthb / 0x100000000)
			event.param1 = lasthb - (event["param0"] * 0x100000000)
		end

	elseif t == "shutdown" then
		-- shutdown info persisted on last shutdown, because there's no direct way to discern
		-- app shutdown from OS shutdown. So, we catch the last one before actual reboot
		-- (want to show os shutdown, not app)
		local lastsd = persist.load("lastshutdown")
		if lastsd then
			if lastsd.time >= transport.reftime then
				-- use shutdown timestamp if valid
				event.timestamp = lastsd.time
			end
			event.param0 = lastsd.reason
		else
			-- no shutdown info - no event
			event = nil
		end
	elseif t == "gpsfix" then
		local qmap = {[0]=0x3000, [1]=0x0000, [2]=0x1000}
	    event.param0 = devicetree.get('system.gps.fix')
		if event.param0 then
			-- Pack the status word
			-- assume no antenna fault
			-- assume 2D fix (not supported)
			-- assume PDOPC Ok (not supported)
			event.param1 = devicetree.get('system.gps.sat_cnt') +							-- NSIV
							(math.min(3, devicetree.get('system.gps.sat_cnt')) * 64) +	-- NSU
							(qmap[devicetree.get('system.gps.qual')] or qmap[0])			-- FD
		else
			-- no GPS available
			event = nil
		end

-- The remaining parameters are filtered via the telemetry filtering mechanism
-- so that the reporting frequency may be configured. The data for these is
-- generally polled rathar than triggered on change.

	elseif t == "wanlinktxbytes" then
		event = wanstate.getCellTxEvent()
		
		if event then
			local tmp = { [t] = event.param1 }
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		end

	elseif t == "wanlinkrxbytes" then
		event = wanstate.getCellRxEvent()
		
		if event then
			local tmp = { [t] = event.param1 }
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		end

	elseif t == "gpslocation" then
		-- Read GPS coords, convert to 9.23 format
		local gpsdata = devicetree.get({"system.gps.latitude", "system.gps.longitude"})
		if gpsdata then
			event.param0 = encodeGps(gpsdata['system.gps.latitude'])
			event.param1 = encodeGps(gpsdata['system.gps.longitude'])

			-- Run thru filter to reduce data usage
			local tmp = { [t] = {latitude = gpsdata['system.gps.latitude'], longitude = gpsdata['system.gps.longitude']} }
			telemetry:filter(tmp)
			if tmp[t] == nil or (gpsdata['system.gps.latitude'] == 0 and gpsdata['system.gps.longitude'] == 0) then
				event = nil
			end
		else
			-- no GPS available
			event = nil
		end

	elseif t == "gpsspeed" then
		-- Convert to M/H, filter via telemetry filter
	    event.param0 = devicetree.get('system.gps.speed') * 1000
		if event.param0 then
			local tmp = { [t] = event.param0 }
			telemetry:filter(tmp)
			if tmp[t] == nil then
				event = nil
			end
		else
			-- no GPS available
			event = nil
		end

	-- set of accelerometer start events
	elseif vbistartparms[t] then
		-- Split the event data string into individual parameters
		-- Use the data from the event - presumably this is queued so
		-- we won't miss events.
		local params = split(v)
		event.param0 = tonumber(params[1])	-- 1st reading past threshold
		event.param1 = tonumber(params[2])	-- event counter
		-- Run thru telmetry filter
		-- Note that ext threshold (always triggers) is used, and
		-- min rate will never be respected, because this is not polled.
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	-- set of accelerometer end events
	elseif vbiendparms[t] then
		-- Split the event data string into individual parameters
		-- Use the data from the event - presumably this is queued so
		-- we won't miss events.
		local params = split(v)
		event.param0 = tonumber(params[2])	-- peak accel
		event.param1 = tonumber(params[4])	-- event counter
		event.extra = "duration=" .. params[1] .. ",average=" .. params[3]

		-- Run thru telmetry filter
		-- Note that ext threshold (always triggers) is used, and
		-- min rate will never be respected, because this is not polled.
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	elseif t == "dreckon_sensor_calibration" then
		event.param0 = tonumber(devicetree.get(msciid.STS_DRECKON_SENSOR_CALIBRATED))
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	elseif t == "dreckon_enabled" then
		event.param0 = tonumber(devicetree.get(msciid.CFG_DRECKON_ENABLE))
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	elseif t == "vpn_tunnel_status" then
		event.param0 = VPNInfo[v].num
		event.param1 = VPNStates[devicetree.get(v)]
		if event.param1 == null then
			event.param1 = VPNStates["default"]
		end
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	elseif t == "vpn_profile_id_map" then
		event.param0 = VPNInfo[v].num
		event.extra = "profile=" .. VPNInfo[v].name
		local tmp = { [t] = 0 }
		telemetry:filter(tmp)
		if tmp[t] == nil then
			event = nil
		end
	else
		--unhandled event
		event = nil
	end

	return event
end

-- Convert IP address as dotted quad string to integer
local function ipToInt (addr)
	local num = 0
	if type(addr) == 'string' then
		addr:gsub("%d+", function(s) num = num * 256 + tonumber(s) end)
	end
	return num
end

local function generateevent(t, v)
    local content = getevent(t, v)
	if content then
		transport:send(content)
	end
end

------------------------------------------------------------------
-- sendProprietaryTelemetryData -
--
-- Proprietary telemetry data are batched and periodically sent.
-- If it's time to send, the collected proprietary data are formatted
-- as a DELS event per datum and sent off to AMM. Data cache is then
-- cleared and publication timestamp is remembered.
--
------------------------------------------------------------------
local function sendProprietaryTelemetryData()

    if BnB.get_CustomTelemData_enable() == 1 then
        if next(TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE) ~= nil then

            local now = transport:systime()
            if now >= (TELEM_CFG_FILE_SIGNAL_LAST_PUBLISH + BnB.get_CustomTelemData_max()) then

                -- Can send now without violating reporting throttle period
                for k,v in pairs(TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE) do
                    local event = {}
                    event.messageid = delseventmap.custom_can_data.messageid
                    event.extra     = "<"..k..">=<"..v..">"
                    log(APP_NAME, "DEBUG", "AMMER-184 - sendProprietaryTelemetryData() calling transport:send(event)")
                    transport:send(event)
                end

                TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE = { }
                TELEM_CFG_FILE_SIGNAL_LAST_PUBLISH  = transport:systime()
            end
        end
    end
end

------------------------------------------------------------------
-- throttleProprietaryTelemetryReporting -
--
-- Every second, if there is proprietary telemetry data cached to send,
-- call sendProprietaryTelemetryData() to try sending it out.
--
------------------------------------------------------------------
local function throttleProprietaryTelemetryReporting()

    while true do
        if next(TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE) ~= nil then
            sendProprietaryTelemetryData()
        end
        sched.wait(1)
    end
end

------------------------------------------------------------------
-- handleProprietaryTelemetryMsciid -
--
-- Process CSM notification for monitored proprietary telemetry data.
-- Save processed data to cache.
-- Call sendProprietaryTelemetryData() to try to send it out.
--
------------------------------------------------------------------
local function handleProprietaryTelemetryMsciid(vars)

    for k,v in pairs(vars) do
        log(APP_NAME, "DEBUG", "handleProprietaryTelemetryMsciid(): "..k.."="..v)

        local name = TELEM_CFG_FILE_SIGNAL_NAME_TBL[k]
        if name then
            TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE[name] = v
            --log(APP_NAME, "DEBUG", "AMMER-184 - handleProprietaryTelemetryMsciid(): create event for... "..name)
        end
    end

    if next(TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE) ~= nil then
        sendProprietaryTelemetryData()
    end
end

------------------------------------------------------------------
-- handleCfgFileDiscovery -
--
-- Discover which proprietary telemetry configuration file is active
-- and register for CSM notifications on the proprietary telemetry
-- data MSCIIDs.
--
------------------------------------------------------------------
local function handleCfgFileDiscovery()

    if (TELEM_CFG_FILE_TITLE and TELEM_CFG_FILE_REVISION) then
        --log(APP_NAME, "DEBUG", "AMMER-184 - TITLE/REVISION = "..TELEM_CFG_FILE_TITLE.."/"..TELEM_CFG_FILE_REVISION)

        -- If new file set has been installed, cleanup any leftover stuff
        TELEM_CFG_FILE_SIGNAL_NAME_TBL = { }
        TELEM_CFG_FILE_SIGNAL_UNIT_TBL = { }
        for k,v in pairs(TELEM_CFG_FILE_SIGNAL_REGISTERED_USERID_TBL) do

            devicetree.unregister(v)
            sched.wait(0.2)
        end
        TELEM_CFG_FILE_SIGNAL_REGISTERED_USERID_TBL = { }

        -- Build up list of candidate files to check
        local fileList = { }
        local numFiles = 0
        local attr = nil

        -- Is this gateway running 4.13.0+ ALEOS code?
        local activeTelemetryCfgFile = "/tmp/activeTelemetryCfg.txt" -- ALEOS Release 4.13.0+ CanMgr will dump a copy here for AMMER|AVTA to retrieve
        attr = lfs.attributes(activeTelemetryCfgFile)
        if attr == nil then
            -- Nope.
            --log(APP_NAME, "DEBUG", "AMMER-184 - "..activeTelemetryCfgFile.." does not exist")
        else
            -- Yep.
            numFiles = numFiles + 1
            fileList[numFiles] = activeTelemetryCfgFile
        end

        -- Was telemetry configuration file installed via the AVTC application?
        local avtcDir = "/mnt/user/readyagent/apps/AVTC/vehiclecfg/"
        attr = nil
        attr = lfs.attributes(avtcDir)
        if attr == nil then
            -- Nope.
            --log(APP_NAME, "DEBUG", "AMMER-184 - "..avtcDir.." does not exist")
        else
            if attr.mode == "directory" then
                -- Yep.
                for file in lfs.dir(avtcDir) do
                    if file ~= "." and file ~= ".." then
                        attr = lfs.attributes(avtcDir..file)
                        if attr.mode ~= "directory" then
                            numFiles  = numFiles + 1
                            fileList[numFiles] = avtcDir..file
                        end
                    end
                end
            else
                -- How can this be???
                log(APP_NAME, "ERROR", avtcDir.." is not a directory!")
            end
        end

        -- Go through candidate list until file with matching title/revision found
        if numFiles < 0 then
            log(APP_NAME, "ERROR", "CAN Mgr advertises Title/Revision but no Telemetry Cfg Files exist!")
            return
        end

        local noMatchFound = true
        local tbl = nil

        for i = 1, numFiles do

            -- Read whole file into a string
            local f = io.open(fileList[i], "rb")
            local content = f:read("*all")
            f:close()

            -- Check file for matching title/revision
            tbl = json.decode(content)
            --log(APP_NAME, "DEBUG", "AMMER-184 - checking file "..fileList[i].." - TITLE/REVISION = "..tbl.VTConfig.title.."/"..tbl.VTConfig.revision)

            if (TELEM_CFG_FILE_TITLE    == tbl.VTConfig.title and
                TELEM_CFG_FILE_REVISION == tbl.VTConfig.revision) then

                noMatchFound = false
                break -- BREAK out of FOR loop; title+revision matched!
            end
        end

        if noMatchFound then

            log(APP_NAME, "ERROR", "No match found for telemetry cfg file title/revision = "..TELEM_CFG_FILE_TITLE.."/"..TELEM_CFG_FILE_REVISION)
            return
        end

        -- Go through contents of 'tbl'
        for i = 1, #tbl.VTConfig.messages do

            for j = 1, #tbl.VTConfig.messages[i].signals do

                -- Proprietary MSCIID range = 59101 to 59150
                if tbl.VTConfig.messages[i].signals[j].itemId >= 59101 and
                   tbl.VTConfig.messages[i].signals[j].itemId <= 59150 then

                    if tbl.VTConfig.messages[i].signals[j].name then

                        local msciid = "aleos."..tbl.VTConfig.messages[i].signals[j].itemId
                        local signalName = string.gsub(tbl.VTConfig.messages[i].signals[j].name, " ", "_")
                        -- Replaced <space> characters in signal name with <underscore> character.

                        -- Register newly discovered "proprietary" MSCIIDs with CSM
                        TELEM_CFG_FILE_SIGNAL_REGISTERED_USERID_TBL[msciid] = devicetree.register(msciid, handleProprietaryTelemetryMsciid)
                        -- Read intial value
                        local initialValue = devicetree.get(msciid)
                        if initialValue ~= "" then
                            -- This MSCIID has indeed been written to before.
                            TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE[signalName] = initialValue
                        end
                        sched.wait(0.2)

                        -- Track and report these signals (proprietary MSCIIDs fall outside BBSE supported items)
                        TELEM_CFG_FILE_SIGNAL_NAME_TBL[msciid] = signalName

                        if tbl.VTConfig.messages[i].signals[j].units then
                            TELEM_CFG_FILE_SIGNAL_UNIT_TBL[msciid] = tbl.VTConfig.messages[i].signals[j].units
                        end
                    end
                end

                --[[ log(APP_NAME, "DEBUG", "AMMER-184 -- msciId = "..tbl.VTConfig.messages[i].signals[j].itemId)
                if tbl.VTConfig.messages[i].signals[j].units then
                    log(APP_NAME, "DEBUG", "AMMER-184 --- units = "..tbl.VTConfig.messages[i].signals[j].units)
                end
                if tbl.VTConfig.messages[i].signals[j].name then
                    log(APP_NAME, "DEBUG", "AMMER-184 ---- name = "..tbl.VTConfig.messages[i].signals[j].name)
                end ]]
            end
        end

        if next(TELEM_CFG_FILE_SIGNAL_SAMPLES_CACHE) ~= nil then
               sendProprietaryTelemetryData()
        end

        --[[ log(APP_NAME, "DEBUG", "AMMER-184 --- TITLE/REVISION MATCH FOUND")
        for k, v in pairs(TELEM_CFG_FILE_SIGNAL_NAME_TBL) do
            log(APP_NAME, "DEBUG", "AMMER-184 ----- Proprietary MSCIID = "..k)
            log(APP_NAME, "DEBUG", "AMMER-184 ------------------- name = "..v)
            if TELEM_CFG_FILE_SIGNAL_UNIT_TBL[k] then
                log(APP_NAME, "DEBUG", "AMMER-184 ------------------- unit = "..TELEM_CFG_FILE_SIGNAL_UNIT_TBL[k])
            end
        end ]]
    end
end

local function msciidtoevent(vars)
	for k,v in pairs(vars) do
		log(APP_NAME, "DEBUG", "msciidtoevent: "..k.."="..v)
	end
    if vars[msciid.STS_TIME_VALID] then
		transport:send(nil)
	end
    if vars["system.gps.fix"] or
		vars["system.gps.qual"] or
	   (vars["system.gps.sat_cnt"] and vars["system.gps.sat_cnt"] <= 4) then
    	generateevent("gpsfix")
    end
    if vars["system.gps.speed"] then
    	generateevent("gpsspeed")
    end
	if vars[msciid.STS_DRECKON_SENSOR_CALIBRATED] then
		generateevent("dreckon_sensor_calibration")
	end

	-- check list of vehicle behaviour events
	-- Check for start events first to ensure that start
	-- always preceeds stop (when they appear simultanious)
	for evt,dtname in pairs(vbistartparms) do
		if vars[dtname] then
			generateevent(evt, vars[dtname])
		end
	end

	for evt,dtname in pairs(vbiendparms) do
		if vars[dtname] then
			generateevent(evt, vars[dtname])
		end
	end

	-- check for individual VPN status changes
	for evt,vpninfo in pairs(VPNInfo) do
		if vars[evt] then
			generateevent("vpn_tunnel_status", evt)
		end
	end

    if vars[msciid.STS_TELEM_CFG_FILE_TITLE] then
        TELEM_CFG_FILE_TITLE = vars[msciid.STS_TELEM_CFG_FILE_TITLE]
        --log(APP_NAME, "DEBUG", "AMMER-184 - MSCIID_STS_TELEM_CFG_FILE_TITLE just set to "..TELEM_CFG_FILE_TITLE)
        handleCfgFileDiscovery()
    end

    if vars[msciid.STS_TELEM_CFG_FILE_REVISION] then
        TELEM_CFG_FILE_REVISION = vars[msciid.STS_TELEM_CFG_FILE_REVISION]
        --log(APP_NAME, "DEBUG", "AMMER-184 - MSCIID_STS_TELEM_CFG_FILE_REVISION just set to "..TELEM_CFG_FILE_REVISION)
        handleCfgFileDiscovery()
    end

    if BnB.get_CustomTelemData_enable() == 1 then
        if tonumber(vars[msciid.STS_TELEM_SCAN_TOOL_STATUS]) then
            TELEM_SCAN_TOOL_STATUS = tonumber(vars[msciid.STS_TELEM_SCAN_TOOL_STATUS])
            local event = {}
            event.messageid = delseventmap.custom_can_data.messageid
            event.extra     = "<scantool_status>=<"..TELEM_SCAN_TOOL_STATUS..">"
            transport:send(event)
        end
    end

    return "ok"
end

-- Send an update of WAN conditions to AMM. This is called
-- when a change in network status is detected. WAN link state
-- generation always looks at entire WAN state, not just changes.
local WanUpdateFilter = {}
local function updateWanEvents()
	local links = wanstate.extractWanLinks()
	local events = wanstate.createWanLinkEvents(links)
	for i, event in ipairs(events) do
		-- Filter out redundant events.
		-- Some events are complementary (WAN up/down and active/notactive),
		-- so the efid field is added to the event when generated; these pairs
		-- have the same efid so that they are compared together instead of as
		-- separate events. The efid essentially creates a slot for each event
		-- to compare that even's previous state to send only on transtition.
		-- Some events add the DPI to the efid, so that the same event applied
		-- to different interface will effectively create a unique slot for that
		-- event/DPI combination. The extra field is not checked, because no
		-- events use that parameter.
		if not WanUpdateFilter[event.efid] or
		WanUpdateFilter[event.efid].messageid ~= event.messageid or
		WanUpdateFilter[event.efid].param0 ~= event.param0 or
		WanUpdateFilter[event.efid].param1 ~= event.param1 then
			WanUpdateFilter[event.efid] = event
			event.efid = nil	-- efid is for internal use only - don't send
			transport:send(event)
		end
	end
end

-- Handle WAN state notification from ALEOS.
-- The action runs after a delay, to prevent multiple notifications from
-- creating a flurry of DELS events (allow time to settle down first).
local WanUpdateTimer
local function dlyUpdateWanEvents(evts)
	if not WanUpdateTimer then
		WanUpdateTimer = timer.once(3,
		function() 
		-- Do work here
		updateWanEvents()
		end )
	else
		timer.rearm(WanUpdateTimer)
	end
end

-- Generate GPIO input events, send as "angle-bracket" stats
local function updateGPInputs()
	local events = {}
	local inputdefs = {
		["input1"] = {stat = "INPUT1", msciid = msciid.IO_DIGITAL_IN1},
		["input2"] = {stat = "INPUT2", msciid = msciid.IO_DIGITAL_IN2},
		["input3"] = {stat = "INPUT3", msciid = msciid.IO_DIGITAL_IN3},
		["input4"] = {stat = "INPUT4", msciid = msciid.IO_DIGITAL_IN4},
		["input5"] = {stat = "INPUT5", msciid = msciid.IO_DIGITAL_IN5},
		["input6"] = {stat = "INPUT6", msciid = msciid.IO_DIGITAL_IN6},
		["input_an1"] = {stat = "INPUT-AN1", msciid = msciid.IO_ANALOG_IN1},
		["input_an2"] = {stat = "INPUT-AN2", msciid = msciid.IO_ANALOG_IN2},
		["input_an3"] = {stat = "INPUT-AN3", msciid = msciid.IO_ANALOG_IN3},
		["input_an4"] = {stat = "INPUT-AN4", msciid = msciid.IO_ANALOG_IN4},
		["input_an5"] = {stat = "INPUT-AN5", msciid = msciid.IO_ANALOG_IN5},
		["input_an6"] = {stat = "INPUT-AN6", msciid = msciid.IO_ANALOG_IN6},
		["input_an7"] = {stat = "INPUT-AN7", msciid = msciid.IO_ANALOG_IN7},
		["input_an8"] = {stat = "INPUT-AN8", msciid = msciid.IO_ANALOG_IN8},
	}

	for name,inp in pairs(inputdefs) do
		-- read input states. Non-existant inputs will read null,
		-- so they will not create an entry (not be generated)
		events[name] = tonumber(devicetree.get(inp.msciid))
	end
	telemetry:filter(events)
	for evt,val in pairs(events) do
		transport:send {messageid=1, extra="<"..inputdefs[evt].stat..">=<"..val..">"}
	end
end

-- Generate management tunnel status events
	local function updateMgmtTunStats()
		-- default tunnel down event
		local events = {tun_state=0, tun_address=0}

		-- Get tunnel info from ifconfig (mgmt tunnel hardcoded as tun1)
		local tuninfo = io.popen("/sbin/ifconfig tun1 2>/dev/null | grep addr:")
		if tuninfo then
			str = tuninfo:read()
			if type(str) == "string" then
				local tunaddr = string.match(str, "%d+%.%d+%.%d+%.%d+")
				if type(tunaddr) == "string" then
					events.tun_state = 1
					events.tun_address = ipToInt(tunaddr)
				end
			end
			tuninfo:close()
		end
		telemetry:filter(events)
		for evt,val in pairs(events) do
			transport:send {messageid=delseventmap[evt].messageid, param0=val}
		end
	end

-- Register for change notifications in response to changing conditions
local function registertoaleos()
	local registerIds = {}	-- table of registration IDs
	local parms = {
		msciid.STS_TIME_VALID,
		msciid.STS_DRECKON_SENSOR_CALIBRATED,
        msciid.STS_TELEM_CFG_FILE_TITLE,
        msciid.STS_TELEM_CFG_FILE_REVISION,
		"system.gps.fix",
		"system.gps.sat_cnt",
		msciid.STS_TELEM_SCAN_TOOL_STATUS,
		msciid.VPN1_STATE,
		msciid.VPN2_STATE,
		msciid.VPN3_STATE,
		msciid.VPN4_STATE,
		msciid.VPN5_STATE,
	}
	for k,v in pairs(vbistartparms) do
		table.insert(parms, v)
	end
	for k,v in pairs(vbiendparms) do
		table.insert(parms, v)
	end

	-- Due to a bug in ALEOS, registering all of the required notifications overflows
	-- the CSM queue. The work-around here slows down the registration process so that
	-- this does not occur.
--	devicetree.register(parms, msciidtoevent)
	for i=1, #parms do table.insert(registerIds, devicetree.register(parms[i], msciidtoevent)) sched.wait(0.2) end

--	devicetree.register(wanstate.msciids, dlyUpdateWanEvents)
	for i=1, #wanstate.msciids do table.insert(registerIds, devicetree.register(wanstate.msciids[i], dlyUpdateWanEvents)) sched.wait(0.2) end

	return registerIds
end

local function checkstartupdown()
	-- Only send boot event one per boot (in case app re-starts).
	-- If flag file (in ramdisk, gets cleared at boot) is present,
	-- startup has already been processed (AMMER is restarting), so bypass.
	-- If flag file not present this is first time since boot, so create it and process.
	local f = io.open(bootflagfile, "r")
	if not f then
		f = io.open(bootflagfile, "w")
		io.close(f)

		local uncleanshutdown = persist.load("lastheartbeat")

		if uncleanshutdown then
			generateevent("uncleanshutdown")
		else
			generateevent("shutdown")
		end
		persist.save("lastshutdown", nil)

		generateevent("startup")
		generateevent("bootignitionstate")
		generateevent("dreckon_sensor_calibration")
		generateevent("dreckon_enabled")
	else
		io.close(f)
	end

    generateevent("appinfo")
    generateevent("caninfo")
	generateevent("rap_id")

	-- send initial individula VPN status
	for evt,vpninfo in pairs(VPNInfo) do
		generateevent("vpn_profile_id_map", evt)
		generateevent("vpn_tunnel_status", evt)
	end
end

------------------------------------------------------------------
-- format_telemetry_msg - package telemetry data into DELS event
-- 
-- Inputs:
--	data		- table of parameter/value pairs
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Passed to temeletry constructor, this recieves telemetry data
--	and reformats/packages into DELS events
------------------------------------------------------------------
local function format_telemetry_msg(data)
	for k,v in pairs(data) do
		if k == "ignition_state" then
			-- send ignition state telemetry event as a DELS event.
			local evt = {}
			evt.messageid = delseventmap.ignitionstate.messageid
			evt.param0 = v
			evt.param1 = tonumber(devicetree.get(IGNITION_SENSE_TIME))
			transport:send(evt)
		elseif DELSparams[k] ~= nil then
			local extra = "MOI="..DEVICE_ID..",MessageString="..tostring(DELSparams[k].id).." "..DELSparams[k].format.." "..tostring(v)
			local evt = {}
			log(APP_NAME, "DETAIL", 'DELS telemetry data: "'..extra..'"')
			evt.messageid = 0
			evt.param0 = 1002
			evt.extra = extra
			transport:send(evt)
		else
			if k == "error" and v == "OBDII disconnected" then
				-- Streamer disconnect sent as battery voltage message
				telemetry:trigger("battery_voltage")
				transport:send({messageid=0, param0=1002, extra="MOI="..DEVICE_ID..",MessageString=83 S DISCONNECTED"})
			elseif k == "info" and v == "OBDII connected" then
				-- ignore Streamer connect
			elseif k == "error" then
				transport:inform("error", v)
			elseif k == "warning" then
				transport:inform("warning", v)
			elseif k == "info" then
				transport:inform("info", v)
			else
				log(APP_NAME, "INFO", "Unhandled telemetry event %s = %s", k, v)
			end
			-- log(APP_NAME, "INFO", "Raw telemetry data: "..tostring(k).."="..tostring(v))
		end
	end
end

-- ----------------------------------------------------------------------------
-- PROCESSES
-- ----------------------------------------------------------------------------
------------------------------------------------------------------
-- main - Main App Process 
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
------------------------------------------------------------------
local function main()
    assert(devicetree.init())
    assert(airvantage.init())
	
	local serial_port

	log(APP_NAME, "INFO", APP_NAME.." Started")
	--print("Executing from: "..os.getenv("PWD"))

	SW_VERSION = devicetree.get("system.sw_info.fw_ver")
	DEVICE_ID = devicetree.get("system.hw_info.global_id")

	-- Determing which line to monitor for ignition input
	-- If MP70 telemetry enabled and using I/O for ignition
	-- then monitor I/O, else monitor ignition.
	-- (If telemetry is disabled, use ignition, because that's the preferred wiring)
	-- (Non-4.9.0 or non-MP70 won't get these options enabled)
	IGNITION_SENSE_LINE = "system.io.ignition_in"
	IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IGNITION_TIMER

	-- This should work for at least the new few versions of ALEOS. After that, everything will be different
	if string.match( SW_VERSION, "^4.1[1-9]" )  then
		CANBUS_ENABLED = true
		local can_cfg = devicetree.get( msciid.CFG_TELEM_CANBUS_ENABLED_IGN_SOURCE )
		if can_cfg == "0" then
			CANBUS_ENABLED = false
			IGNITION_SENSE_LINE = "system.io.ignition_in"
			IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IGNITION_TIMER
		elseif can_cfg == "1" then
			IGNITION_SENSE_LINE = "system.io.ignition_in"
			IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IGNITION_TIMER
		elseif can_cfg == "2" then
			IGNITION_SENSE_LINE = "system.io.gpio.1"
			IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IO_LINE_TIMER
		elseif can_cfg == "3" then
			-- Ignition line is not wired - if we're powered, assume it's on 
			IGNITION_SENSE_LINE = "always_on"
			IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IO_LINE_TIMER
		end
	elseif string.match( SW_VERSION, "^4.9" ) then
		if devicetree.get(msciid.CFG_TELEM_CANBUS_ENABLED) == "1" then
			CANBUS_ENABLED = true
			if devicetree.get(msciid.CFG_TELEM_IGN_SOURCE) == "2" then
				IGNITION_SENSE_LINE = "system.io.gpio.1"
				IGNITION_SENSE_TIME = msciid.CFG_LOW_PWR_IO_LINE_TIMER
			end
		else
			CANBUS_ENABLED = false
		end
	end

	log( APP_NAME, "INFO", "Detected ALEOS " .. SW_VERSION .. ", using " .. IGNITION_SENSE_LINE .. " as ignition trigger" )

	transport = _m3da(APP_NAME, "eventrpt", "event")
	BnB = _BnB(APP_NAME)

	local bbem = _bbem("127.0.0.1", 0xbb5e)
	if BnB:connect(bbem) then
		log(APP_NAME, "INFO", "Using BBSE Streamer interface (detected)")
	--[[
		WTF alert; can only read MSCIID 10120 properly when it contains 0;
		will get <nil> back when MSCIID 10120 contains 1.
		Need to investigate that and resolve.
		Work around for now.
	--]]
	elseif devicetree.get(msciid.CFG_AAF_RESERVE_SERIAL_PORT) == 0 then -- Note: Check passes for 0, not "0".
		log(APP_NAME, "INFO", "Fallback to BBSE Streamer interface (serial port not reserved)")
	else
		bbem:close()
		bbem=nil
		serial_port = serial.open("/dev/ttyS0", {baudRate=115200,flowControl="none"})
		BnB:connect(serial_port)
		log(APP_NAME, "INFO", "Using serial Streamer interface")
	end

	telemetry = _telemetry(APP_NAME, BnB, format_telemetry_msg)
	
	if not transport or not BnB or not telemetry then
		log(APP_NAME, "ERROR", "Cannot initialize")
		return
	end

	telemetry.options = persist.load(CONFIG_ASSET_NAME)
	if not telemetry.options then
		telemetry.options = {}
	end
	BnB:get_default_options(telemetry.options)
	-- load and activate system options
	for soptname, soptval in pairs(default_system_options) do
		-- check for nil because false is a valid default value
		-- false means it's saved outside options. The (false) value
		-- is save there so routines know to act on it, but don't
		-- initialize via handler
		if telemetry.options[soptname] == nil then
			telemetry.options[soptname] = soptval.value
		end
		if soptval.set_handler and telemetry.options[soptname] then
			soptval.set_handler(soptname, telemetry.options[soptname])
		end
	end

	-- Set up the asset for configuration
	local st,err = airvantage.init()
	if not st then
		log(APP_NAME, "ERROR", "Cannot initialize av lib %s", err)
		return
	end

	configasset = airvantage.newAsset(CONFIG_ASSET_NAME)
	configasset.tree.downlink = recieve_config
	configasset.tree.uplink = request_config
	configasset:start()

    transport:run()

	-- Stop ALEOS from spamming the log with M3DA junk (AMMER gets blamed,
	-- but it's really an AAF issue). Don't override an existing setting.
	-- Each of these modules spams the log. Lua console command tt.logfix sets
	-- another level (eg tt.logfix "INFO"). Log level nil removes settings.
	-- Note: boot is required for this to take effect.
	local waitboot = false
	if not devicetree.get("config.log.moduleslevel.M3DA-SESSION") then
		log(APP_NAME, "INFO", "Setting M3DA-SESSION log level to NONE")
		devicetree.set("config.log.moduleslevel.M3DA-SESSION", "NONE")
		waitboot=true
	end
	if not devicetree.get("config.log.moduleslevel.SRVCON") then
		log(APP_NAME, "INFO", "Setting SRVCON log level to NONE")
		devicetree.set("config.log.moduleslevel.SRVCON", "NONE")
		waitboot=true
	end
	if not devicetree.get("config.log.moduleslevel.DATAMGR") then
		log(APP_NAME, "INFO", "Setting DATAMGR log level to NONE")
		devicetree.set("config.log.moduleslevel.DATAMGR", "NONE")
		waitboot=true
	end

	if waitboot then
		-- Set up just enough to be able to get a message out to AMM
		transport:inform("warning", "AMMER Initial install - must reboot GW to activate")

		-- debugging stuff
		tt = {}
		tt.devicetree = devicetree
		tt.wanstate = wanstate
		tt.transport = transport

		-- abort main - other routines will keep app active
		return
	end

	-- trap signal to get app shutdown notification
	psignal.signal("SIGTERM",
		function()
			local now = transport:systime()
			log(APP_NAME, "WARNING", APP_NAME.." SIGTERM Received ...")
			-- remove unclean shutdown marker lastheartbeat
			persist.save("lastheartbeat", nil)
			
			-- save shutdown time and reason. On first execution will generate shutdown event
			-- using this timestamp and code. Note that ALEOS sometimes sends this signal
			-- more than once, so ignore all but first (reset at boot).
			if not persist.load("lastshutdown") then
				persist.save("lastshutdown", {time = now, reason = 0} )
			end
			
			-- signal rest of app to stop
			sched.signal(APP_NAME, "EXIT")
			log(APP_NAME, "WARNING", APP_NAME.." Exiting ...")
		end
	)

	-- generate startup/shutdown events
	checkstartupdown()
	
	local last_heartbeat = 0

	-- Set up WAN state module before first use, then trigger for initial update
	wanstate.init(devicetree,
		function(message) log(APP_NAME, "DEBUG", message) end
	)
	
	-- Register for various operational events
    local registerIds = registertoaleos()
	
	-- Get most up to date WAN status
	updateWanEvents()

	-- For debug access via telnet console
	if DEBUG then
		tt = {
			devicetree = devicetree,
			persist = persist,
			getevent = getevent,
			generateevent = generateevent,
			split = split,
			transport = transport,
			telemetry = telemetry,
			wanstate = wanstate,
			updateWanEvents= updateWanEvents,
			dlyUpdateWanEvents = dlyUpdateWanEvents,
			WanUpdateFilter = WanUpdateFilter,
			delseventmap = delseventmap,
			configasset = configasset,
			vbistartparms = vbistartparms,
			vbiendparms = vbiendparms,
			default_system_options = default_system_options,
			recieve_config = recieve_config,
			request_config = request_config,
			msciidtoevent = msciidtoevent,
			msciid = msciid,
			registerIds = registerIds,
			updateGPInputs = updateGPInputs,
			updateMgmtTunStats = updateMgmtTunStats,
			-- list of options/values (eg tt.setopt {heartbeat="enable=1, min=120", sys_log_level="DETAIL"} )
			setopt = function(opts) recieve_config(configasset, opts, "uplink") end,
			getopt = function() return get_config(configasset, "", "downlink") end,
			logfix = function(level)	-- set squelched logs to different level
				devicetree.set("config.log.moduleslevel.M3DA-SESSION", level)
				devicetree.set("config.log.moduleslevel.SRVCON", level)
				devicetree.set("config.log.moduleslevel.DATAMGR", level)
				print("Don't forget to reboot for changes to take effect.")
			end,
		}
	end

	telemetry:start()

    local title = devicetree.get(msciid.STS_TELEM_CFG_FILE_TITLE)
    if (title ~= "NOTSET") then
        TELEM_CFG_FILE_TITLE = title
        --log(APP_NAME, "DEBUG", "AMMER-184 - MSCIID_STS_TELEM_CFG_FILE_TITLE already set to "..TELEM_CFG_FILE_TITLE)
    end
    local revision = devicetree.get(msciid.STS_TELEM_CFG_FILE_REVISION)
    if (revision ~= "NOTSET") then
        TELEM_CFG_FILE_REVISION = revision
        --log(APP_NAME, "DEBUG", "AMMER-184 - MSCIID_STS_TELEM_CFG_FILE_REVISION already set to "..TELEM_CFG_FILE_REVISION)
    end

    TELEM_SCAN_TOOL_STATUS = tonumber(devicetree.get(msciid.STS_TELEM_SCAN_TOOL_STATUS))
    if TELEM_SCAN_TOOL_STATUS then
        if BnB.get_CustomTelemData_enable() == 1 then
            local event = {}
            event.messageid = delseventmap.custom_can_data.messageid
            event.extra     = "<scantool_status>=<"..TELEM_SCAN_TOOL_STATUS..">"
            transport:send(event)
        end
    end

    handleCfgFileDiscovery()
    sched.run(throttleProprietaryTelemetryReporting)

    -- Main loop - sit and poll gps and stuff
	local tuncntr = 10
	while true do
		
		-- The following are polled periodically, but are sent
		-- according to telemetry filter rules
		generateevent("cpu_load")
		generateevent("main_supply")
		generateevent("board_temperature")
		generateevent("heartbeat")
		generateevent("wanlinktxbytes")
		generateevent("wanlinkrxbytes")
		generateevent("wan_sig_ext")

		updateGPInputs()

		tuncntr = tuncntr + 1
		if tuncntr > 10 then
			-- don't need to check tunnel stats every second
			updateMgmtTunStats()
			tuncntr = 0
		end

		-- poll for hardware-based ignition state if no streamer present
		if not telemetry:isPresent() then
			generateevent("ignitionstate")
		end
		local gpsfix = devicetree.get('system.gps.fix')
		if gpsfix and gpsfix ~= 0 then
			generateevent("gpslocation")
			generateevent("gpsspeed")
		end

		local sig = sched.wait(APP_NAME, {"EXIT", 1})
		if sig == "EXIT" then
			break
		end
	end

	print("Exiting " .. APP_NAME .. " cleaning up ...")

	-- Execute orderly shutdown
	telemetry:stop()

	if bbem then
		bbem:close()
	else
		serial_port:close()
	end

	transport:stop()

    -- go thru registerIds and unregister them all.
    for k,v in pairs(registerIds) do
        devicetree.unregister(v)
        sched.wait(0.1)
    end
    registerIds = {}

    for k,v in pairs(TELEM_CFG_FILE_SIGNAL_REGISTERED_USERID_TBL) do
        devicetree.unregister(v)
        sched.wait(0.1)
    end
    TELEM_CFG_FILE_SIGNAL_REGISTERED_USERID_TBL = {}

	configasset:close()

	os.exit()
end

------------------------------------------------------------------
-- main - Debug shell 
-- 
-- Inputs:
--
-- Outputs:
--
-- Returns:
--
-- Notes
--	Start a telnet server on port 1234
--	Once this program is started, you can start a Lua VM through telnet
--	using the following command: telnet localhost 1234
------------------------------------------------------------------
if DEBUG then
	local function run_server()
	local shell = require 'shell.telnet'
	
	shell.init {
		address     = '127.0.0.1',
		port        = 1234,
		editmode    = "edit",
		historysize = 100 }
	end

	sched.run(run_server)
end

sched.run(main)
sched.loop()
